from simpletrainer.common import oprator as op
from simpletrainer.common.attrs_extend import define, field
from simpletrainer.common.default_settings import DefaultSettings
from simpletrainer.core import (
    AttrsComponent,
    BaseComponent,
    StatefulAttrsComponent,
    Trainer,
    TrainerConfig,
    after,
    before,
    on,
)
from simpletrainer.version import __version__

__all__ = [
    'Trainer',
    'TrainerConfig',
    'DefaultSettings',
    'BaseComponent',
    'AttrsComponent',
    'StatefulAttrsComponent',
    'define',
    'field',
    'after',
    'before',
    'on',
    'op',
    '__version__',
]



import socket
import traceback
from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from simpletrainer import Trainer

DATE_FORMAT = '%Y-%m-%d %H:%M'


class NotificationType(str, Enum):
    START = 'start'
    FINISH = 'finish'
    CRASH = 'crash'


@dataclass
class NotificationInfo:
    content: str
    type_: NotificationType


def start_info(trainer: Trainer) -> NotificationInfo:
    host_name = socket.gethostname()
    date = datetime.now().strftime(DATE_FORMAT)
    contents = [
        f'Machine: {host_name}',
        f'Model: {trainer.model.__class__.__name__}',
        f'Experiment: {trainer.experiment_name}',
        f'Start: {date}',
    ]
    return NotificationInfo(
        content='\n'.join(contents),
        type_=NotificationType.START,
    )


def finish_info(trainer: Trainer) -> NotificationInfo:
    host_name = socket.gethostname()
    date = datetime.now().strftime(DATE_FORMAT)

    contents = [
        f'Machine: {host_name}',
        f'Model: {trainer.model.__class__.__name__}',
        f'Experiment: {trainer.experiment_name}',
        f'End: {date}',
        'Train Metrics:',
        str(trainer.train_metrics_history),
    ]
    if trainer.do_eval:
        contents.append('Validation Metrics:')
        contents.append(str(trainer.valid_metrics_history))
    return NotificationInfo(
        content='\n'.join(contents),
        type_=NotificationType.FINISH,
    )


def crash_info(trainer: Trainer) -> NotificationInfo:
    host_name = socket.gethostname()
    date = datetime.now().strftime(DATE_FORMAT)

    contents = [
        f'Machine: {host_name}',
        f'Model: {trainer.model.__class__.__name__}',
        f'Experiment: {trainer.experiment_name}',
        f'Crash: {date}',
        "Here's the error:",
        f'trainer.exception\n\n',
        'Traceback:',
        str(traceback.format_exc()),
    ]
    return NotificationInfo(
        content='\n'.join(contents),
        type_=NotificationType.CRASH,
    )



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
from typing import Dict, Optional

import torch
import torch.nn as nn
from attrs import asdict
from rich.align import Align
from rich.columns import Columns
from rich.console import Console, Group
from rich.live import Live
from rich.logging import RichHandler
from rich.panel import Panel
from rich.pretty import Pretty
from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.rule import Rule
from rich.table import Column, Table
from rich.text import Text
from torch.optim import Optimizer
from torchinfo import ModelStatistics

import simpletrainer.utils.torch as torch_utils
from simpletrainer import BaseComponent, Trainer, after, before, on
from simpletrainer.common.types import MetricDict
from simpletrainer.core.hook import HookCollection, Preposition, TrainerEvent
from simpletrainer.utils.common import pretty_repr

console = Console()
logger = logging.getLogger(__name__)


def convert_dict_to_table(data: dict, title: str):
    column_names = [
        'name',
        'value',
    ]
    table = Table(
        *[Column('[bold]' + col) for col in column_names],
        title=title,
        title_style='i b magenta',
    )
    for k, v in data.items():
        table.add_row(
            Text(k, style='b'),
            Pretty(v),
        )
    return table


class RichLiveObj:
    def __init__(self):
        self.live_objs = {}

    def get_metrics_section(self) -> list:
        stage_metrics = self.live_objs.get('stage_metrics')
        if stage_metrics is None:
            return []
        else:
            panels = [Panel.fit(f'[b]{k}[/b]:\n[green]{v:.3f}[/green]') for k, v in stage_metrics.items()]
            return panels

    def __rich_console__(self, console, options):
        objs = [
            Rule('Trainer Progress'),
            Text('\n'),
            Columns([self.live_objs['progress']] + self.get_metrics_section()),
        ]
        if self.live_objs.get('epoch_metrics'):
            objs += [
                Text('\n'),
                self.live_objs.get('epoch_metrics'),
            ]
        yield from objs


class RichProgressBar(BaseComponent):
    only_main_process = True
    tags = ['progress_bar']
    train_task_id: TaskID
    valid_task_id: TaskID

    def __init__(self):
        self.progress = Progress(
            '{task.description}',
            BarColumn(),
            TextColumn('{task.fields[steps]}{task.fields[total_steps]}'),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
            TextColumn('[progress.percentage]{task.percentage:>3.0f}%'),
            console=console,
        )
        self.epoch_metrics_table = Table(title='[magenta]Epoch Metrics[/magenta]')
        self.rich_live_obj = RichLiveObj()
        self._live = Live(console=console)
        self._rich_handler_added = False

    def _prepare_rich_handler(self):
        self.rich_handler = RichHandler(console=console)
        logging.root.addHandler(self.rich_handler)
        self.removed_handlers = []
        for handler in logging.root.handlers:
            if isinstance(handler, logging.StreamHandler):
                self.removed_handlers.append(handler)
                logging.root.removeHandler(handler)
        logger.debug(f'Remove handlers: {self.removed_handlers}')

    @staticmethod
    def get_total_steps_str(num_batches_per_epoch: Optional[int] = None):
        if num_batches_per_epoch is None:
            total_steps = ''
        else:
            total_steps = f'/{num_batches_per_epoch}'
        return total_steps

    def add_task(self, task_name: str, total_steps: Optional[int] = None):
        task_id = self.progress.add_task(
            f'[green]{task_name.capitalize()}[/green]',
            total=total_steps or 0,
            steps=0,
            start=False,
            total_steps=self.get_total_steps_str(total_steps),
        )
        return task_id

    @on(Trainer.EVENT.START)
    def fit_progress(self, trainer: Trainer):
        self.train_task_id = self.add_task('train', trainer.train_data_info.num_batches_per_epoch)
        if trainer.do_eval:
            self.valid_task_id = self.add_task('valid', trainer.valid_data_info.num_batches_per_epoch)  # type: ignore
        self._prepare_rich_handler()
        self._live.start()
        self._live.update(self.rich_live_obj)

    @after(Trainer.prepare_for_stage)
    def set_initial(self, trainer: Trainer):
        if trainer.state.num_batches:
            self.progress.reset(self.train_task_id, completed=trainer.state.num_batches)

    @after(Trainer.run_batch)
    def advance_batch(self, trainer: Trainer):
        task_id = self.train_task_id if trainer.in_train_stage else self.valid_task_id
        self.progress.update(task_id, steps=trainer.state.num_batches, advance=1)

    @before(Trainer.run_stage)
    def start_train_stage(self, trainer: Trainer):
        if trainer.in_train_stage:
            self.progress.start_task(self.train_task_id)
        else:
            train_steps = trainer.train_data_info.num_batches_per_epoch
            if train_steps is not None:
                self.progress.update(self.train_task_id, steps=train_steps)
            self.progress.start_task(self.valid_task_id)

        self.rich_live_obj.live_objs['stage_metrics'] = trainer.state.stage_metrics

    @before(Trainer.run_epoch)
    def set_progress(self, trainer: Trainer):
        self.rich_live_obj.live_objs['progress'] = Panel.fit(
            self.progress,  # type: ignore
            title=f'Epoch {trainer.current_epoch}/{trainer.epochs}',
        )

    @after(Trainer.run_epoch)
    def refresh_progress(self, trainer: Trainer):

        self.progress.reset(
            self.train_task_id,
            start=False,
            steps=0,
            total_steps=self.get_total_steps_str(trainer.train_data_info.num_batches_per_epoch),
        )
        if trainer.do_eval:
            self.progress.reset(
                self.valid_task_id,
                start=False,
                steps=0,
                total_steps=self.get_total_steps_str(trainer.valid_data_info.num_batches_per_epoch),  # type: ignore
            )
        self.rich_live_obj.live_objs['progress'].title = f'Epoch {trainer.current_epoch}/{trainer.epochs}'
        self.rich_live_obj.live_objs['epoch_metrics'] = self.get_epoch_metrics_tabel(
            trainer.train_metrics_history,
            trainer.valid_metrics_history,
        )

    @on(Trainer.EVENT.TEARDOWN)
    def stop(self, trainer: Trainer):
        self.progress.stop()
        self._live.stop()
        for handler in logging.root.handlers:
            if handler is self.rich_handler:
                logging.root.removeHandler(handler)
        for handler in self.removed_handlers:
            logging.root.addHandler(handler)

    @staticmethod
    def get_epoch_metrics_tabel(
        train_metrics_list: list[MetricDict],
        valid_metrics_list: list[MetricDict],
    ):
        metric_set = set()
        for metrics in train_metrics_list:
            metric_set |= set(f'train-{i}' for i in metrics)
        for metrics in valid_metrics_list:
            metric_set |= set(f'valid-{i}' for i in metrics)

        all_metrics = sorted(list(metric_set))
        cols = ['epoch'] + all_metrics

        epoch_metrics_table = Table(*cols, title='[magenta]Epoch Metrics[/magenta]')
        for idx in range(len(train_metrics_list)):
            row_values = [Text(str(idx + 1), style='bold')]
            for metric in all_metrics:
                if metric.startswith('train'):
                    value = train_metrics_list[idx][metric[6:]]
                else:
                    value = valid_metrics_list[idx][metric[6:]]
                row_values.append(Text(str(round(value, 3))))
            epoch_metrics_table.add_row(*row_values)
        return epoch_metrics_table

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'


class RichInspect(BaseComponent):
    only_main_process = True
    tags = ['inspect']

    def __init__(self, model_summary_depth: int = 3):
        self.model_summary_depth = model_summary_depth

    @on(Trainer.EVENT.START)
    def show_trainer_info(self, trainer: Trainer):
        named_panels = [
            (
                'Model Info',
                self.get_model_panel(
                    trainer.raw_model,
                    None,
                    trainer.raw_optimizer,
                    trainer.device,
                ),
            ),
            (
                'Optimizer Info',
                self.get_optimizer_panel(trainer.raw_optimizer),
            ),
            (
                'Data Info',
                self.get_data_panel(trainer.train_data_info, trainer.valid_data_info),
            ),
            ('Trainer Info', self.get_trainer_panel(trainer)),
            (
                'Trainer Hook',
                self.get_hook_tabel_panel(trainer.hook_engine.entrypoint_hooks),
            ),
            (
                'Trainer Components',
                self.get_components_panel(trainer._components),
            ),
        ]
        for name, panels in named_panels:
            self.print_section(name, panels)

    @staticmethod
    def get_module_group_set(module, id_group_map: Dict[int, str]):
        group_set = set()
        for _, p in module.named_parameters():
            group = id_group_map.get(id(p))
            if group is not None:
                group_set.add(group)
        return group_set

    @staticmethod
    def get_model_param_panels(model_statistics: ModelStatistics):
        param_info = {
            'Total params': f'[green]{model_statistics.total_params}[/green]',
            'Trainable params': f'[green]{model_statistics.trainable_params}[/green]',
            'Non-trainable params': f'\
            [green]{model_statistics.total_params - model_statistics.trainable_params}[/green]',
        }
        panels = [Panel(f'[b]{k}:\n[/b]{v}') for k, v in param_info.items()]
        return panels

    def get_model_table(
        self,
        model_statistics: ModelStatistics,
        model: nn.Module,
        optimizer: Optimizer,
    ):
        column_names = [
            'Module',
            'Params',
            'Optimizer Group',
            'Output Shape',
            'Trainable',
        ]
        model_table = Table(
            *[Column('[bold]' + col) for col in column_names],
            title=f'{model.__class__.__name__} Architecture',
            title_style='i b magenta',
        )

        parameter_id_group_map = torch_utils.get_parameter_id_group_map(optimizer)
        id_layer_map = {layer.layer_id: layer for layer in model_statistics.summary_list}
        for module in model.modules():
            layer = id_layer_map.get(id(module))
            if layer is None or layer.depth > self.model_summary_depth:
                continue

            if layer.depth == 0:
                module_name = f'[bold green]{layer.class_name}'
            else:
                module_name = '[blue]|  [/blue]' * (layer.depth - 1) + '[blue]└─ [/blue]' + f'[bold green]{layer.class_name}'
            num_params = Pretty(layer.num_params) if layer.num_params else '--'
            _group_set = self.get_module_group_set(module, parameter_id_group_map)
            group = Pretty(_group_set) if _group_set else '--'
            output_size = Pretty(layer.output_size) if layer.output_size else '--'
            trainable = Pretty(layer.trainable_params != 0)
            model_table.add_row(module_name, num_params, group, output_size, trainable)
        return model_table

    @staticmethod
    def get_optimizer_panel(optimizer: Optimizer):
        optimizer_panel = Panel.fit(Pretty(optimizer), title='[b i magenta]Optimizer')
        return [Align.center(Columns([optimizer_panel], padding=(0, 5), equal=True))]

    @staticmethod
    def get_trainer_panel(trainer: Trainer):
        info = trainer.info
        return [
            Align.center(
                Columns(
                    [
                        convert_dict_to_table(asdict(info.accelerator), title='Accelerator'),
                        convert_dict_to_table(info.hyper_params, title='HyperParams'),
                        convert_dict_to_table(asdict(info.init_attributes), title='InitAttributes'),
                    ],
                    padding=(0, 5),
                    equal=True,
                )
            )
        ]

    @staticmethod
    def get_data_panel(train_data_info, valid_data_info):
        panels = [Panel.fit(Pretty(train_data_info), title='Train Data Info')]
        if valid_data_info is not None:
            panels.append(Panel.fit(Pretty(valid_data_info), title='Valid Data Info'))
        return [
            Align.center(
                Columns(
                    panels,
                    padding=(0, 5),
                    equal=True,
                )
            )
        ]

    @staticmethod
    def print_section(section_name: str, section_elements: list) -> None:
        section_elements.insert(0, Rule(f'[b]{section_name}'))
        for paragraph in section_elements:
            if isinstance(paragraph, list):
                for ele in paragraph:
                    console.print(ele)
            else:
                console.print(paragraph)
            console.print('\n')
        console.print('\n')

    def get_model_panel(
        self,
        model: nn.Module,
        input_data,
        optimizer: Optimizer,
        device: torch.device,
    ):
        model_statistics = torch_utils.get_model_info(model, input_data, device=device)
        model_table = self.get_model_table(model_statistics, model, optimizer)
        params_panels = self.get_model_param_panels(model_statistics)
        model_class_name = model.__class__.__name__

        return [
            Align.center(model_table),
            [
                Align.center(
                    Text(
                        f'{model_class_name} Params Statistics',
                        style='b i magenta',
                    )
                ),
                Align.center(Columns(params_panels, equal=True)),
            ],
        ]

    @staticmethod
    def get_components_panel(components):
        return [Align.center(Panel.fit(Group(*(Pretty(component) for component in components))))]

    @staticmethod
    def get_hook_tabel_panel(hooks: dict[str, HookCollection]):
        column_names = [
            'hook',
            'prep',
            'name',
        ]
        hook_table = Table(
            *[Column('[bold]' + col) for col in column_names],
            title='Hooks',
            title_style='i b magenta',
        )
        hook_event_set = set(i.value for i in list(TrainerEvent))
        for entrypoint_name, hook_collection in hooks.items():
            if entrypoint_name in hook_event_set:
                entrypoint_name = f'Event({entrypoint_name})'
            for preposition in Preposition:
                for hook in hook_collection.data[preposition]:
                    hook_table.add_row(
                        Text(entrypoint_name, style='b'),
                        Text(preposition, style='green'),
                        Text(hook.__qualname__),
                    )
        return [Align.center(hook_table)]

    @staticmethod
    def get_hyper_params_panel(hyper_params):
        return [Align.center(Panel(pretty_repr(hyper_params, name='HyperParams')), width=100)]

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'



from pathlib import Path
from typing import Optional, Union

from accelerate import Accelerator
from accelerate.utils import DistributedDataParallelKwargs, PrecisionType
from pydantic import BaseModel, Field, validator

from simpletrainer.common.default_settings import DefaultSettings
from simpletrainer.common.sign_metric import SignMetric
from simpletrainer.common.types import PathOrStr
from simpletrainer.utils.common import pretty_repr, random_experiment_name


class AcceleratorConfig(BaseModel):
    cpu: bool = Field(default=False, description='Whether or not to force the script to execute on CPU.')
    mixed_precision: PrecisionType = PrecisionType.NO
    ddp_kwargs: dict = Field(default_factory=dict)

    def build(self):
        args = []
        if self.ddp_kwargs:
            args.append(DistributedDataParallelKwargs(**self.ddp_kwargs))

        _mixed_precision = self.mixed_precision

        accelerator = Accelerator(
            mixed_precision=_mixed_precision,
            cpu=self.cpu,
            kwargs_handlers=args,
        )
        return accelerator


class TrainerConfig(BaseModel):
    epochs: int = DefaultSettings.epochs
    do_eval: bool = True
    accumulate_grad_batches: int = 1
    core_metric: SignMetric = SignMetric('-', 'loss')
    experiment_name: str = DefaultSettings.experiment_name
    run_name: str = Field(default_factory=lambda: random_experiment_name())
    output_dir: Optional[Path] = None
    checkpoint_for_restore: Optional[PathOrStr] = None
    auto_restore: bool = True
    inspect: bool = False
    progress_bar: Union[str, None] = 'rich'
    accelerator: AcceleratorConfig = Field(default_factory=AcceleratorConfig)
    logger: Union[str, None] = None

    @validator('core_metric')
    def convert_string_to_sign_metric(cls, v):
        if isinstance(v, str):
            return SignMetric.from_str(v)
        elif isinstance(v, SignMetric):
            return v
        raise ValueError(f'Invalid core_metric: {v}')

    def __repr__(self) -> str:
        return pretty_repr(self.dict(), self.__class__.__name__)



import importlib.util


def is_wandb_available():
    return importlib.util.find_spec('wandb') is not None


def is_transformers_available():
    return importlib.util.find_spec('transformers') is not None


def is_torchmetrics_available():
    return importlib.util.find_spec('torchmetrics') is not None


def is_mlflow_available():
    return importlib.util.find_spec('mlflow') is not None



from __future__ import annotations

from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class Stateful(Protocol):
    def state_dict(self) -> dict[str, Any]:
        ...

    def load_state_dict(self, state_dict: dict[str, Any]):
        ...



from __future__ import annotations

from typing import Optional, Union

import torch

from simpletrainer.common.types import HyperParams, MetricDict
from simpletrainer.loggers.base import DeepLearningLogger


class DummyLogger(DeepLearningLogger):
    def log_metrics(self, scalars: dict[str, Union[int, float]], step: Optional[int] = None) -> None:
        pass

    def log_tensors(self, tensors: dict[str, torch.Tensor], step: Optional[int] = None) -> None:
        pass

    def log_summary(self, hyper_params: HyperParams, metrics: MetricDict) -> None:
        pass

    def log_artifact(self, artifact: str) -> None:
        pass



from __future__ import annotations

import os
import warnings
from typing import TYPE_CHECKING, Optional, Union

import torch

from simpletrainer.common.types import HyperParams, MetricDict
from simpletrainer.integrations import is_mlflow_available
from simpletrainer.loggers.base import DeepLearningLogger

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer

if is_mlflow_available():
    import mlflow


class MlflowLogger(DeepLearningLogger):
    def __init__(self, tracking_uri: Optional[str] = None):
        if not is_mlflow_available():
            raise ImportError('wandb is not installed')
        self.tracking_uri = tracking_uri or os.getenv('MLFLOW_TRACKING_URI', None)

        if self.tracking_uri is not None:
            mlflow.set_tracking_uri(self.tracking_uri)

    def with_trainer(self, trainer: 'Trainer') -> None:
        mlflow.set_experiment(trainer.experiment_name)

    def log_metrics(self, scalars: dict[str, Union[int, float]], step: Optional[int] = None) -> None:
        _scalars = {k: float(v) for k, v in scalars.items()}
        mlflow.log_metrics(_scalars, step=step)

    def log_tensors(self, tensors: dict[str, torch.Tensor], step: Optional[int] = None) -> None:
        warnings.warn('MlflowLogger does not support tensor logging')

    def log_summary(self, hyper_params: HyperParams, metrics: MetricDict) -> None:
        mlflow.log_params(hyper_params)
        mlflow.log_metrics(metrics)

    def log_artifact(self, artifact: str) -> None:
        mlflow.log_artifact(artifact)



from pydantic import BaseSettings


class SimpleTrainergDefaultSettings(BaseSettings):
    experiment_name: str = 'untitled_experiment'
    epochs: int = 3
    seed: int = 42
    core_metric: str = '-loss'
    logger: str = 'dummy'
    log_file_name: str = 'trainer.log'
    best_model_file_name: str = 'best_model.pt'
    best_model_state_file_name: str = 'best_model_state.pt'
    metrics_file_name = 'metrics.json'
    realtime_metric_prefix: str = 'rt_'
    model_state_file_name: str = 'pytorch_model.bin'
    trainer_info_json_file_name: str = 'trainer_info.json'
    trainer_state_json_file_name: str = 'trainer_state.json'
    checkpoints_root_dir_name: str = 'checkpoints'
    checkpoints_register_book_file_name: str = 'checkpoints_register_book.txt'
    pred_key: str = 'logits'
    target_key: str = 'labels'
    metric_round: int = 4
    log_format: str = '[%(levelname)s] %(asctime)s %(module)s: %(message)s'
    date_format = '%Y-%m-%d %H:%M'

    class Config:
        env_prefix = 'SIMPLETRAINER_'


DefaultSettings = SimpleTrainergDefaultSettings()



# type: ignore
from typing import Any, Optional, Sequence

from attrs import NOTHING
from attrs import define as attrs_define
from attrs import field as attrs_field


def define(
    maybe_cls=None,
    *,
    these=None,
    only_main_process: bool = False,
    tags: Optional[Sequence[str]] = None,
    repr=None,
    hash=None,
    init=None,
    slots=True,
    frozen=False,
    weakref_slot=True,
    str=False,
    auto_attribs=None,
    kw_only=False,
    cache_hash=False,
    auto_exc=True,
    eq=None,
    order=False,
    auto_detect=True,
    getstate_setstate=None,
    on_setattr=None,
    field_transformer=None,
    match_args=True,
):
    def new_func(cls):
        cls = attrs_define(
            maybe_cls=cls,
            these=these,
            repr=repr,
            hash=hash,
            init=init,
            slots=slots,
            frozen=frozen,
            weakref_slot=weakref_slot,
            str=str,
            auto_attribs=auto_attribs,
            kw_only=kw_only,
            cache_hash=cache_hash,
            auto_exc=auto_exc,
            eq=eq,
            order=order,
            auto_detect=auto_detect,
            getstate_setstate=getstate_setstate,
            on_setattr=on_setattr,
            field_transformer=field_transformer,
            match_args=match_args,
        )
        setattr(cls, 'only_main_process', only_main_process)
        setattr(cls, 'tags', tags or [])
        return cls

    if maybe_cls is None:
        return new_func
    else:
        return new_func(maybe_cls)


def field(
    *,
    default=NOTHING,
    export: bool = False,
    hyper_param: bool = False,
    save: bool = False,
    validator=None,
    repr=True,
    hash=None,
    init=True,
    metadata=None,
    converter=None,
    factory=None,
    kw_only=False,
    eq=None,
    order=None,
    on_setattr=None,
) -> Any:
    new_metadata = metadata or {}
    new_metadata.update(simpletrainer_export=export, simpletrainer_hyper_param=hyper_param, simpletrainer_save=save)
    return attrs_field(
        default=default,
        validator=validator,
        repr=repr,
        hash=hash,
        init=init,
        metadata=new_metadata,
        converter=converter,
        factory=factory,
        kw_only=kw_only,
        eq=eq,
        order=order,
        on_setattr=on_setattr,
    )



from __future__ import annotations

import re
from dataclasses import asdict, dataclass
from typing import Optional, Sized, TypeVar

import torch
import torchinfo
from accelerate.utils.random import set_seed
from torch.utils.data import DataLoader

from simpletrainer.utils.common import pretty_repr

T = TypeVar('T')
set_seed = set_seed


@dataclass
class DataInfo:
    batch_size: int
    batch_size_per_device: int
    num_sampels: Optional[int]
    num_batches_per_epoch: Optional[int]

    def __repr__(self) -> str:
        return pretty_repr(asdict(self), self.__class__.__name__)


def get_batch_size_from_dataloader(dataloader: DataLoader) -> int:
    if dataloader.batch_size is None:
        try:
            return dataloader.batch_sampler.batch_size  # type: ignore
        except AttributeError:
            raise ValueError(
                'Can not get batch size from dataloader, does not support `BatchSampler` with varying batch size yet.'
            )
    else:
        return dataloader.batch_size


def get_num_samples_from_dataloader(dataloader: DataLoader) -> Optional[int]:
    if isinstance(dataloader.dataset, Sized):
        return len(dataloader.dataset)
    elif isinstance(dataloader.sampler, Sized):
        return len(dataloader.sampler)
    else:
        sampler = getattr(dataloader.batch_sampler, 'sampler')
        if isinstance(sampler, Sized):
            return len(sampler)
        else:
            return


def get_data_info(dataloader: DataLoader, world_size: int = 1) -> DataInfo:
    num_samples = get_num_samples_from_dataloader(dataloader)

    try:
        num_batches_per_epoch = len(dataloader)
    except:
        num_batches_per_epoch = None

    batch_size_per_device = get_batch_size_from_dataloader(dataloader)
    batch_size = batch_size_per_device * world_size

    return DataInfo(
        batch_size=batch_size,
        batch_size_per_device=batch_size_per_device,
        num_sampels=num_samples,
        num_batches_per_epoch=num_batches_per_epoch,
    )


def get_model_info(
    model: torch.nn.Module,
    input_data: Optional[torchinfo.torchinfo.INPUT_DATA_TYPE] = None,
    device: Optional[torch.device] = None,
) -> torchinfo.ModelStatistics:
    try:
        model_statistics = torchinfo.summary(model, input_data=input_data, verbose=0, device=device)
    except Exception:
        model_statistics = torchinfo.summary(model, verbose=0, device=device)
    return model_statistics


def get_parameter_id_group_map(
    optimizer: torch.optim.Optimizer,
) -> dict[int, str]:
    parameter_id_group_map = {}
    for group, params in enumerate(optimizer.param_groups):
        for param in params['params']:
            parameter_id_group_map[id(param)] = str(group)
    return parameter_id_group_map


def get_params_with_pattern(model: torch.nn.Module, pattern: re.Pattern):
    params = []
    for name, param in model.named_parameters():
        if pattern.search(name):
            params.append(param)
    return params


def get_module_learning_rate_summary(module: torch.nn.Module, optimizer: torch.optim.Optimizer):
    lr_dict: dict[str, float] = {}
    names = {param: name for name, param in module.named_parameters()}
    for group in optimizer.param_groups:
        if 'lr' not in group:
            continue
        lr = group['lr']
        for param in group['params']:
            if param.requires_grad:
                lr_dict[names[param]] = lr
            else:
                lr_dict[names[param]] = 0.0
    return lr_dict


def get_module_parameter_summary(model: torch.nn.Module):
    parameter_mean: dict[str, float] = {}
    parameter_std: dict[str, float] = {}

    for name, param in model.named_parameters():
        if param.data.numel() > 0:
            parameter_mean[name] = float(param.data.mean().item())
        if param.data.numel() > 1:
            parameter_std[name] = float(param.data.std().item())
    return parameter_mean, parameter_std


def get_module_gradient_summary(model: torch.nn.Module):
    gradient_mean: dict[str, float] = {}
    gradient_std: dict[str, float] = {}

    for name, param in model.named_parameters():
        if param.grad is not None:
            if param.grad.is_sparse:
                grad_data = param.grad.data._values()
            else:
                grad_data = param.grad.data

            # skip empty gradients
            if torch.prod(torch.tensor(grad_data.shape)).item() > 0:
                gradient_mean[name] = float(grad_data.mean().item())
                if grad_data.numel() > 1:
                    gradient_std[name] = float(grad_data.std().item())
    return gradient_mean, gradient_std



# -*- coding: utf-8 -*-
from typing import Optional

from simpletrainer import AttrsComponent, Trainer, before, define, field


@define
class GradClip(AttrsComponent):
    grad_norm: Optional[float] = field(default=1.0, hyper_param=True)
    grad_value: Optional[float] = field(default=None, hyper_param=True)
    only_main_process = False

    def __attrs_post_init__(self):
        if self.grad_norm is None and self.grad_value is None:
            raise ValueError('Either grad_norm or grad_value must be specified.')

    @before(Trainer.step, try_last=True)
    def clip_grad(self, trainer: Trainer):
        if self.grad_norm is not None:
            trainer.accelerator.clip_grad_norm_(trainer.model.parameters(), self.grad_norm)

        if self.grad_value is not None:
            trainer.accelerator.clip_grad_value_(trainer.model.parameters(), self.grad_value)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

import torch

from simpletrainer.common.types import HyperParams, MetricDict
from simpletrainer.integrations import is_wandb_available
from simpletrainer.loggers.base import DeepLearningLogger

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer

if is_wandb_available():
    import wandb


class WandbLogger(DeepLearningLogger):
    def __init__(self):
        if not is_wandb_available():
            raise ImportError('wandb is not installed')

    def with_trainer(self, trainer: Trainer) -> None:
        wandb.init(
            project=trainer.experiment_name,
            name=trainer.run_name,
            reinit=True,
        )

    def log_scalars(self, scalars: dict[str, Union[int, float]], step: Optional[int] = None) -> None:
        for name, scalar in scalars.items():
            wandb.log({name: scalar}, step=step)

    def log_tensors(self, tensors: dict[str, torch.Tensor], step: Optional[int] = None) -> None:
        for name, tensor in tensors.items():
            tensor = tensor.cpu().data.numpy().flatten()
            wandb.log({name: wandb.Histogram(tensor)}, step=step)

    def log_summary(self, hyper_params: HyperParams, metrics: MetricDict) -> None:
        wandb.config.update(hyper_params)
        for k, v in metrics.items():
            wandb.run.summary[k] = v  # type: ignore

    def log_artifact(self, artifact: str) -> None:
        wandb.log_artifact(artifact)

    def teardown(self) -> None:
        wandb.finish()



from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

from attrs import asdict, define

from simpletrainer.common.types import PathOrStr

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer


@define
class TrainerInitAttributes:
    epochs: int
    do_eval: bool
    accumulate_grad_batches: int
    core_metric: str
    experiment_name: str
    run_name: str
    output_dir: str
    auto_restore: bool


@define
class AcceleratorInfo:
    mixed_precision: str
    num_processes: int
    device: str
    distributed_type: str


@define
class TrainerInfo:
    init_attributes: TrainerInitAttributes
    accelerator: AcceleratorInfo
    logger: str
    components: list[str]
    hyper_params: dict

    train_metrics: list[dict]
    valid_metrics: list[dict]

    model: str
    optimizer: str
    lr_scheduler: str

    @classmethod
    def from_trainer(cls, trainer: 'Trainer'):
        if hasattr(trainer, 'model'):
            model = trainer.raw_model.__class__.__name__
        else:
            model = ''

        if hasattr(trainer, 'optimizer'):
            optimizer = trainer.raw_optimizer.__class__.__name__
        else:
            optimizer = ''

        if hasattr(trainer, 'lr_scheduler'):
            lr_scheduler = trainer.lr_scheduler.__class__.__name__
        else:
            lr_scheduler = ''

        init_attributes = TrainerInitAttributes(
            epochs=trainer.epochs,
            do_eval=trainer.do_eval,
            accumulate_grad_batches=trainer.accumulate_grad_batches,
            core_metric=str(trainer.core_metric),
            experiment_name=trainer.experiment_name,
            run_name=trainer.run_name,
            output_dir=str(trainer.output_dir),
            auto_restore=trainer.auto_restore,
        )

        acclerator = AcceleratorInfo(
            mixed_precision=str(trainer.accelerator.state.mixed_precision),
            num_processes=trainer.accelerator.state.num_processes,
            device=str(trainer.accelerator.state.device),
            distributed_type=str(trainer.accelerator.state.distributed_type),
        )

        return cls(
            init_attributes=init_attributes,
            accelerator=acclerator,
            logger=trainer.logger.__class__.__name__,
            components=[i.__class__.__name__ for i in trainer._components],
            hyper_params=trainer.hyper_params,
            train_metrics=trainer.state.train_metrics_history,
            valid_metrics=trainer.state.valid_metrics_history,
            model=model,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
        )

    def to_json(self, json_file: PathOrStr) -> None:
        Path(json_file).write_text(json.dumps(asdict(self)))



from simpletrainer.components.terminal.rich_terminal import RichInspect, RichProgressBar
from simpletrainer.components.terminal.tqdm_progress_bar import TqdmProgressBar

__all__ = ['RichInspect', 'RichProgressBar', 'TqdmProgressBar']



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional, Union

import torch
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from torch.utils.data import DataLoader

from simpletrainer.common.default_settings import DefaultSettings
from simpletrainer.common.protocols import Stateful
from simpletrainer.common.sign_metric import SignMetric
from simpletrainer.common.types import (
    Batch,
    BatchOutput,
    HyperParams,
    LRScheduler,
    MetricDict,
    PathOrStr,
)
from simpletrainer.core.configs import TrainerConfig
from simpletrainer.core.hook import TrainerEvent, TrainerHookEngine, entrypoint
from simpletrainer.core.info import TrainerInfo
from simpletrainer.core.mixins import AcceleratorMixin, TrainerStateMixin
from simpletrainer.core.state import TrainerState
from simpletrainer.loggers import DeepLearningLogger, DeepLearningLoggerRegistry
from simpletrainer.utils.common import random_experiment_name, smartget
from simpletrainer.utils.torch import get_data_info

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from simpletrainer.core.component import BaseComponent


class Trainer(TrainerStateMixin, AcceleratorMixin):
    EVENT = TrainerEvent

    model: torch.nn.Module
    optimizer: torch.optim.Optimizer
    train_dataloader: DataLoader
    valid_dataloader: Optional[DataLoader]
    lr_scheduler: Optional[LRScheduler]

    def __init__(
        self,
        epochs: int = DefaultSettings.epochs,
        do_eval: bool = True,
        accumulate_grad_batches: int = 1,
        core_metric: SignMetric = SignMetric.from_str(DefaultSettings.core_metric),
        experiment_name: str = DefaultSettings.experiment_name,
        run_name: Optional[str] = None,
        output_dir: Optional[Path] = None,
        auto_restore: bool = True,
        checkpoint_for_restore: Optional[PathOrStr] = None,
        accelerator: Optional[Accelerator] = None,
        logger: Optional[DeepLearningLogger] = None,
        components: Iterable['BaseComponent'] = tuple(),
    ) -> None:
        self.epochs = epochs
        self.do_eval = do_eval
        self.accumulate_grad_batches = accumulate_grad_batches
        self.core_metric = core_metric
        self.experiment_name = experiment_name
        self.run_name = run_name or random_experiment_name()
        self.output_dir = output_dir or (Path(self.experiment_name) / self.run_name)
        self.auto_restore = auto_restore
        self.checkpoint_for_restore = checkpoint_for_restore
        self.accelerator = accelerator or Accelerator()
        self.logger = logger or DeepLearningLoggerRegistry[DefaultSettings.logger]()
        self._components: list['BaseComponent'] = []
        self.exports: dict[str, Any] = {}
        self.state = TrainerState()

        self.hook_engine = TrainerHookEngine(self)
        self.register_components(components)

    @classmethod
    def from_config(cls, config: TrainerConfig) -> 'Trainer':
        from simpletrainer.components import (
            RichInspect,
            RichProgressBar,
            TqdmProgressBar,
        )

        accelerator = config.accelerator.build()
        if config.logger:
            logger = DeepLearningLoggerRegistry[config.logger]()
        else:
            logger = None

        components = []
        if config.inspect:
            components.append(RichInspect())
        if config.progress_bar == 'rich':
            components.append(RichProgressBar())
        elif config.progress_bar == 'tqdm':
            components.append(TqdmProgressBar())

        return cls(
            accelerator=accelerator,
            logger=logger,
            epochs=config.epochs,
            do_eval=config.do_eval,
            accumulate_grad_batches=config.accumulate_grad_batches,
            core_metric=config.core_metric,
            experiment_name=config.experiment_name,
            run_name=config.run_name,
            output_dir=config.output_dir,
            auto_restore=config.auto_restore,
            checkpoint_for_restore=config.checkpoint_for_restore,
            components=components,
        )

    def train(
        self,
        model: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        train_dataloader: DataLoader,
        valid_dataloader: Optional[DataLoader] = None,
        lr_scheduler: Optional[LRScheduler] = None,
    ) -> None:
        self.model = model
        self.optimizer = optimizer
        self.train_dataloader = train_dataloader
        self.valid_dataloader = valid_dataloader
        self.lr_scheduler = lr_scheduler

        self.prepare()
        self.try_restore_checkpoint(self.checkpoint_for_restore)

        try:
            self.hook_engine.on(TrainerEvent.START)
            self.start_loop()
            self.hook_engine.on(TrainerEvent.FINISH)
        except (Exception, KeyboardInterrupt) as e:
            self.exception = e
            self.crash()
            raise
        finally:
            self.teardown()

    def prepare(self) -> None:
        self._accelerate_prepare()

        if self.valid_dataloader is None:
            self.do_eval = False
        self.train_data_info = get_data_info(self.train_dataloader, self.accelerator.num_processes)
        if self.valid_dataloader is None:
            self.valid_data_info = None
        else:
            self.valid_data_info = get_data_info(self.valid_dataloader, self.accelerator.num_processes)

        self.output_dir.mkdir(exist_ok=True, parents=True)
        self.logger.post_init(self)
        for component in self._components:
            component.post_init_with_trainer(self)
        self.info.to_json(self.output_dir / DefaultSettings.trainer_info_json_file_name)

    def try_restore_checkpoint(self, checkpoint: Optional[PathOrStr] = None) -> None:
        if checkpoint is not None:
            self.load(checkpoint)
        elif self.auto_restore:
            if not self.checkpoint_dir.exists():
                return
            latest_checkpoint = max((d for d in self.checkpoint_dir.iterdir() if d.is_dir()), key=lambda x: x.stat().st_mtime)
            self.load(latest_checkpoint)

    def teardown(self) -> None:
        self.logger.teardown()
        self.info.to_json(self.output_dir / DefaultSettings.trainer_info_json_file_name)
        self.hook_engine.on(TrainerEvent.TEARDOWN)

    def crash(self) -> None:
        if isinstance(self.exception, KeyboardInterrupt):
            self.save('on-keyboard-interrupt')
        self.hook_engine.on(TrainerEvent.CRASH)

    @entrypoint
    def start_loop(self) -> None:
        while not self.should_stop_loop():
            self.run_epoch()
            self.advance_epoch()

    @entrypoint
    def should_stop_loop(self) -> bool:
        return self.current_epoch > self.epochs

    @entrypoint
    def run_epoch(self) -> None:
        if self.state.should_restore:
            self.run_restore_epoch()
        else:
            self.prepare_for_stage(train=True)
            self.run_stage(train=True)
            if self.do_eval:
                self.prepare_for_stage(train=False)
                self.run_stage(train=False)

    @entrypoint
    def run_restore_epoch(self) -> None:
        tn = self.run_train_stage.__name__
        vn = self.run_valid_stage.__name__

        if tn not in self.state.checkpoint_entrypoint_stack and vn not in self.state.checkpoint_entrypoint_stack:
            self.state.checkpoint_entrypoint_stack = []
            return

        if tn in self.state.checkpoint_entrypoint_stack:
            self.prepare_for_stage(train=True)
            self.run_stage(train=True)

        if self.do_eval:
            logger.debug(f'[epoch {self.current_epoch}] valid stage start')
            self.prepare_for_stage(train=False)
            self.run_stage(train=False)

    @entrypoint
    def run_stage(self, train: bool) -> None:
        if train:
            self.run_train_stage()
        else:
            self.run_valid_stage()

    @entrypoint
    def prepare_for_stage(self, train: bool) -> None:
        self.model.train(train)
        batch_iterator = iter(self.train_dataloader) if train else iter(self.valid_dataloader)   # type: ignore
        if self.state.should_restore:
            self.state.restore(train, batch_iterator)
        else:
            self.state.prepare(train, batch_iterator)

    @entrypoint
    def run_train_stage(self) -> None:
        while batch := self.generate_batch():
            self.set_batch(batch)
            self.run_batch()

            if self.should_step():
                self.step()

        stage_metrics = self.collect_stage_metrics()
        stage_metrics = {k: round(v, DefaultSettings.metric_round) for k, v in stage_metrics.items()}
        logger.debug(f'[epoch {self.current_epoch}] train stage metrics: {stage_metrics}')
        self.train_metrics_history.append(stage_metrics)

    @entrypoint
    def generate_batch(self) -> Optional[Batch]:
        if self.state.batch_iterator is None:
            raise ValueError('Batch iterator is not initialized')
        try:
            return next(self.state.batch_iterator)
        except StopIteration:
            return

    @entrypoint
    def set_batch(self, batch: Batch) -> None:
        self.state.batch = batch

    @entrypoint
    def run_valid_stage(self) -> None:
        while batch := self.generate_batch():
            self.set_batch(batch)
            self.run_batch()

        stage_metrics = self.collect_stage_metrics()
        stage_metrics = {k: round(v, DefaultSettings.metric_round) for k, v in stage_metrics.items()}
        logger.debug(f'[epoch {self.current_epoch}] valid stage metrics: {stage_metrics}')
        self.valid_metrics_history.append(stage_metrics)

    @entrypoint
    def collect_stage_metrics(self) -> MetricDict:
        return self.state.stage_metrics

    @entrypoint
    def should_step(self) -> bool:
        return (self.state.num_batches % self.accumulate_grad_batches) == 0

    @entrypoint
    def run_batch(self) -> None:
        self.advance_batch()
        if self.in_train_stage:
            self.run_train_batch()
        else:
            self.run_valid_batch()

    @entrypoint
    def run_train_batch(self) -> None:
        self.forward()
        self.backward()

    @entrypoint
    def run_valid_batch(self) -> None:
        with torch.no_grad():
            self.forward()

    @entrypoint
    def forward(self) -> None:
        batch_output: BatchOutput = self.model(**self.state.batch)
        self.set_batch_output(batch_output)
        self.set_loss()

    @entrypoint
    def backward(self) -> None:
        self.accelerator.backward(self.state.loss)

    @entrypoint
    def set_batch_output(self, batch_output: BatchOutput) -> None:
        self.state.batch_output = batch_output

    @entrypoint
    def set_loss(self) -> None:
        loss = smartget(self.state.batch_output, 'loss')
        self.state.set_loss(loss, self.accumulate_grad_batches)

    @entrypoint
    def step(self) -> None:
        self.advance_step()
        self.optimizer.step()
        self.optimizer.zero_grad()

    @property
    def components(self):
        return self._components

    @property
    def hyper_params(self) -> HyperParams:
        hyper_params: HyperParams = {
            'epochs': self.epochs,
            'mixed_precision': self.mixed_precision,
            'accumulate_grad_batches': self.accumulate_grad_batches,
        }
        for component in self._components:
            hyper_params.update(**component.hyper_params)
        return hyper_params

    @property
    def num_steps_per_epoch(self) -> Optional[int]:
        if self.train_data_info.num_batches_per_epoch is None:
            return
        return self.train_data_info.num_batches_per_epoch // self.accumulate_grad_batches

    @property
    def total_steps(self):
        if self.num_steps_per_epoch is None:
            return
        return self.num_steps_per_epoch * self.epochs

    @property
    def raw_model(self) -> torch.nn.Module:
        return self.accelerator.unwrap_model(self.model)

    @property
    def raw_optimizer(self) -> torch.optim.Optimizer:
        if isinstance(self.optimizer, AcceleratedOptimizer):
            return self.optimizer.optimizer  # type: ignore
        else:
            return self.optimizer

    @property
    def checkpoint_dir(self) -> Path:
        return self.output_dir / DefaultSettings.checkpoints_root_dir_name

    @property
    def info(self) -> TrainerInfo:
        return TrainerInfo.from_trainer(self)

    def register_components(self, components: Iterable['BaseComponent']) -> None:
        for component in components:
            self.hook_engine.register(component)

    def log(
        self, data: Mapping[str, Union[float, torch.Tensor]], step: Optional[int] = None, prefix: Optional[str] = None
    ) -> None:
        if prefix is None:
            prefix = 'train' if self.in_train_stage else 'validation'

        if step is None:
            step = self.cum_steps

        scalars = {k: v for k, v in data.items() if isinstance(v, (int, float))}
        if scalars:
            scalars = self.logger.add_prefix(prefix, scalars)
            self.logger.log_metrics(scalars, step=step)

        tensors = {k: v for k, v in data.items() if isinstance(v, torch.Tensor)}
        if tensors:
            tensors = self.logger.add_prefix(prefix, tensors)
            self.logger.log_tensors(tensors, step=step)

    def save(self, checkpoint_name_or_dir: PathOrStr) -> None:
        is_checkpoint_name = isinstance(checkpoint_name_or_dir, str)
        if is_checkpoint_name:
            save_dir = self.output_dir / DefaultSettings.checkpoints_root_dir_name / checkpoint_name_or_dir
        else:
            if checkpoint_name_or_dir.is_dir():
                save_dir = checkpoint_name_or_dir
            else:
                raise ValueError(f'checkpoint_name_or_dir must be a directory, got {checkpoint_name_or_dir}')

        self.accelerator.save_state(str(save_dir))
        self.state.to_json(save_dir / DefaultSettings.trainer_state_json_file_name)

    def load(self, checkpoint_name_or_dir: PathOrStr) -> None:
        if isinstance(checkpoint_name_or_dir, Path):
            if checkpoint_name_or_dir.is_dir():
                load_dir = checkpoint_name_or_dir
            else:
                raise ValueError(f'checkpoint_name_or_dir must be a directory, got {checkpoint_name_or_dir}')
        else:
            load_dir = self.output_dir / DefaultSettings.checkpoints_root_dir_name / checkpoint_name_or_dir

        self.accelerator.load_state(str(load_dir))
        self.state = self.state.from_json(load_dir / DefaultSettings.trainer_state_json_file_name)
        self.state.checkpoint_entrypoint_stack = self.state.entrypoint_stack
        self.state.entrypoint_stack = []
        self.accelerator.wait_for_everyone()

    def _accelerate_prepare(self) -> None:
        self.model, self.optimizer, self.train_dataloader, self.lr_scheduler = self.accelerator.prepare(  # type: ignore
            self.model, self.optimizer, self.train_dataloader, self.lr_scheduler
        )  # type: ignore
        if self.valid_dataloader is not None:
            self.valid_dataloader = self.accelerator.prepare(self.valid_dataloader)  # type: ignore
        for component in self._components:
            if isinstance(component, Stateful):
                self.accelerator.register_for_checkpointing(component)

    def __getitem__(self, key: str):
        return self.exports[key]

    def __setitem__(self, key: str, item: Any):
        self.exports[key] = item



from __future__ import annotations

import json
import math
from itertools import islice
from pathlib import Path
from typing import Any, Iterator, Optional, Union

import torch
from attrs import define, field, fields

from simpletrainer.common.types import Batch, MetricDict, PathOrStr


def _init_metrics():
    return {'loss': 0.0}


@define
class TrainerState:
    # LoopState
    current_epoch: int = 1
    cum_train_batches: int = 0
    cum_valid_batches: int = 0
    cum_steps: int = 0
    train_metrics_history: list[MetricDict] = field(factory=list)
    valid_metrics_history: list[MetricDict] = field(factory=list)
    entrypoint_stack: list[str] = field(factory=list)
    exception: Optional[Union[Exception, KeyboardInterrupt]] = None

    # StageState
    num_steps: int = 0
    num_batches: int = 0
    stage_metrics: dict[str, float] = field(factory=_init_metrics)
    in_train_stage: bool = True

    batch_iterator: Iterator[Batch] = field(init=False)
    loss: torch.Tensor = field(init=False)
    batch: Batch = field(init=False)
    batch_output: Any = field(init=False)

    # For Restore
    checkpoint_entrypoint_stack: list[str] = field(factory=list)

    @property
    def should_restore(self) -> bool:
        return bool(self.checkpoint_entrypoint_stack)

    def set_loss(self, loss: torch.Tensor, accumulate_grad_batches: int) -> None:
        self.loss = loss / accumulate_grad_batches
        loss_item = loss.item()
        if math.isnan(loss_item):
            return
        self.stage_metrics['loss'] += (loss_item - self.stage_metrics['loss']) / (self.num_batches)

    def prepare(self, train: bool, batch_iterator: Iterator[Batch]):
        self.reset_stage_state()
        self.in_train_stage = train
        self.batch_iterator = batch_iterator

    def restore(self, train: bool, batch_iterator: Iterator[Batch]):
        if train:
            self.batch_iterator = islice(batch_iterator, self.num_batches, None)
        else:
            self.batch_iterator = batch_iterator
        self.checkpoint_entrypoint_stack = []

    def reset_stage_state(self):
        self.num_steps = 0
        self.num_batches = 0
        self.stage_metrics = _init_metrics()

    @classmethod
    def from_json(cls, json_file: PathOrStr) -> TrainerState:
        return cls(**json.loads(Path(json_file).read_text()))

    def to_json(self, json_file: PathOrStr) -> None:
        init_fields = [field.name for field in fields(self.__class__) if field.init]   # type: ignore
        Path(json_file).write_text(json.dumps({field: getattr(self, field) for field in init_fields}, default=str))



# -*- coding: utf-8 -*-
import re
from typing import Optional, Pattern, Sequence

import torch
from torch.cuda.amp import GradScaler, autocast

from simpletrainer import AttrsComponent, Trainer, after, define, field
from simpletrainer.utils.torch import get_params_with_pattern


@define
class GradPenalty(AttrsComponent):
    pattern: Pattern = field(converter=re.compile, hyper_param=True)
    penalty_weight: float = field(default=1.0, hyper_param=True)

    @after(Trainer.set_loss, try_first=True)
    def add_penalty(self, trainer: Trainer):
        if trainer.in_valid_stage:
            return

        loss = trainer.state.loss
        params = get_params_with_pattern(trainer.model, self.pattern)

        grad_scaler: Optional[GradScaler] = trainer.grad_scaler
        if grad_scaler is not None:
            grad_params = torch.autograd.grad(outputs=grad_scaler.scale(loss), inputs=params, create_graph=True)  # type: ignore
            inv_scale = 1.0 / grad_scaler.get_scale()
            grad_params = [p * inv_scale for p in grad_params]
            with autocast():
                penalty = self.calculate_penalty(grad_params)
        else:
            grad_params = torch.autograd.grad(outputs=loss, inputs=params, create_graph=True)
            penalty = self.calculate_penalty(grad_params)
        trainer.update_stage_metrics(grad_penalty=(penalty * self.penalty_weight).item())
        trainer.state.loss = loss + penalty * self.penalty_weight

    @staticmethod
    def calculate_penalty(grad_params: Sequence[torch.Tensor]):
        if not grad_params:
            raise ValueError('No parameters found')

        grad_norm = torch.tensor(0, device=grad_params[0].device, dtype=grad_params[0].dtype)
        for grad in grad_params:
            grad_norm += grad.pow(2).sum()
        grad_norm = grad_norm.sqrt()
        return grad_norm



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Optional, Union

import torch
from torch.utils.tensorboard.summary import hparams
from torch.utils.tensorboard.writer import SummaryWriter

from simpletrainer.common.types import HyperParams, MetricDict
from simpletrainer.loggers.base import DeepLearningLogger

if TYPE_CHECKING:
    from simpletrainer.core import Trainer


class TensorboardLogger(DeepLearningLogger):
    writer: SummaryWriter

    def post_init(self, trainer: 'Trainer') -> None:
        self.writer = SummaryWriter(str(trainer.output_dir))

    def log_metrics(self, scalars: dict[str, Union[int, float]], step: Optional[int] = None) -> None:
        for name, scalar in scalars.items():
            self.writer.add_scalar(name, scalar, step)

    def log_tensors(self, tensors: dict[str, torch.Tensor], step: Optional[int] = None) -> None:
        for name, tensor in tensors.items():
            tensor = tensor.cpu().data.numpy().flatten()
            self.writer.add_histogram(name, tensor, step)

    def log_summary(self, hyper_params: HyperParams, metrics: MetricDict) -> None:
        metric_dict = {f'hparam/{k}': v for k, v in metrics.items()}
        exp, ssi, sei = hparams(hyper_params, metric_dict)
        self.writer.file_writer.add_summary(exp)  # type: ignore
        self.writer.file_writer.add_summary(ssi)  # type: ignore
        self.writer.file_writer.add_summary(sei)  # type: ignore
        for k, v in metric_dict.items():
            self.writer.add_scalar(k, v)

    def log_artifact(self, artifact: str) -> None:
        warnings.warn('Tensorboard does not support artifact logging')

    def teardown(self) -> None:
        self.writer.close()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, NoReturn, Optional

from attrs import fields

from simpletrainer.common.types import HyperParams

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer


class BaseComponent:
    only_main_process: ClassVar[bool]
    tags: ClassVar[list[str]] = []

    @property
    def hyper_params(self) -> HyperParams:
        return {}

    @property
    def exports(self) -> dict[str, Any]:
        return {}

    def check_compatibility(self, other_component: BaseComponent) -> Optional[NoReturn]:
        return

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        return


class AttrsComponent(BaseComponent):
    @property
    def hyper_params(self) -> HyperParams:
        component_fileds = fields(self.__class__)  # type: ignore
        hyper_param_fields = [f for f in component_fileds if f.metadata.get('simpletrainer_hyper_param', False)]
        return {f.name: getattr(self, f.name) for f in hyper_param_fields}

    @property
    def exports(self) -> dict[str, Any]:
        component_fileds = fields(self.__class__)   # type: ignore
        export_fields = [f for f in component_fileds if f.metadata.get('simpletrainer_export', False)]
        return {f.name: getattr(self, f.name) for f in export_fields}


class StatefulAttrsComponent(AttrsComponent):
    def state_dict(self):
        component_fileds = fields(self.__class__)   # type: ignore
        save_fields = [f for f in component_fileds if f.metadata.get('simpletrainer_save', False)]
        return {f.name: getattr(self, f.name) for f in save_fields}

    def load_state_dict(self, state_dict: dict[str, Any]):
        for name, value in state_dict.items():
            setattr(self, name, value)


class ComponentCheckError(Exception):
    pass



from __future__ import annotations

from typing import Optional, Union

import torch
from accelerate import Accelerator

from simpletrainer.common.types import MetricDict
from simpletrainer.core.state import TrainerState


class AcceleratorMixin:
    accelerator: Accelerator

    @property
    def device(self) -> torch.device:
        return self.accelerator.device

    @property
    def is_main_process(self) -> bool:
        return self.accelerator.is_main_process

    @property
    def grad_scaler(self) -> Optional[torch.cuda.amp.GradScaler]:
        return self.accelerator.scaler

    @property
    def mixed_precision(self) -> str:
        return self.accelerator.mixed_precision


class TrainerStateMixin:
    state: TrainerState

    def advance_batch(self):
        if self.state.in_train_stage:
            self.state.cum_train_batches += 1
        else:
            self.state.cum_valid_batches += 1
        self.state.num_batches += 1

    def advance_step(self):
        self.state.cum_steps += 1
        self.state.num_steps += 1

    def advance_epoch(self):
        self.state.current_epoch += 1

    def update_stage_metrics(self, **metric_dict: float) -> None:
        self.state.stage_metrics.update(metric_dict)

    @property
    def current_epoch(self) -> int:
        return self.state.current_epoch

    @property
    def total_train_batches(self) -> int:
        return self.state.cum_train_batches

    @property
    def total_valid_batches(self) -> int:
        return self.state.cum_valid_batches

    @property
    def cum_steps(self) -> int:
        return self.state.cum_steps

    @property
    def total_steps(self) -> int:
        return self.state.cum_steps

    @property
    def train_metrics_history(self) -> list[MetricDict]:
        return self.state.train_metrics_history

    @property
    def valid_metrics_history(self) -> list[MetricDict]:
        return self.state.valid_metrics_history

    @property
    def latest_train_metrics(self) -> Optional[MetricDict]:
        if self.state.train_metrics_history:
            return self.state.train_metrics_history[-1]

    @property
    def latest_valid_metrics(self) -> Optional[MetricDict]:
        if self.state.valid_metrics_history:
            return self.state.valid_metrics_history[-1]

    @property
    def exception(self) -> Optional[Union[Exception, KeyboardInterrupt]]:
        return self.state.exception

    @exception.setter
    def exception(self, value: Union[Exception, KeyboardInterrupt]):
        self.state.exception = value

    @property
    def in_train_stage(self):
        return self.state.in_train_stage

    @property
    def in_valid_stage(self):
        return not self.state.in_train_stage



# -*- coding: utf-8 -*-
import logging

from simpletrainer import AttrsComponent, DefaultSettings, Trainer, define, on


@define(only_main_process=True)
class FileHandler(AttrsComponent):
    log_file_name: str = DefaultSettings.log_file_name
    level: int = logging.INFO

    def post_init_with_trainer(self, trainer: Trainer):
        log_file = trainer.output_dir / self.log_file_name
        self.file_handler = logging.FileHandler(str(log_file))
        self.file_handler.setLevel(self.level)
        logging.root.addHandler(self.file_handler)

    @on(Trainer.EVENT.TEARDOWN)
    def remove_handler(self, trainer: Trainer):
        if hasattr(self, 'file_handler'):
            logging.root.removeHandler(self.file_handler)



from __future__ import annotations

from typing import Type

from simpletrainer.loggers.base import DeepLearningLogger
from simpletrainer.loggers.dummy import DummyLogger
from simpletrainer.loggers.mlflow import MlflowLogger
from simpletrainer.loggers.tensorboard import TensorboardLogger
from simpletrainer.loggers.wandb import WandbLogger

DeepLearningLoggerRegistry: dict[str, Type[DeepLearningLogger]] = {
    'tensorboard': TensorboardLogger,
    'wandb': WandbLogger,
    'mlflow': MlflowLogger,
    'dummy': DummyLogger,
}



from simpletrainer.core.component import (
    AttrsComponent,
    BaseComponent,
    ComponentCheckError,
    StatefulAttrsComponent,
)
from simpletrainer.core.configs import AcceleratorConfig, TrainerConfig
from simpletrainer.core.hook import after, before, on
from simpletrainer.core.trainer import Trainer

__all__ = [
    'AcceleratorConfig',
    'TrainerConfig',
    'before',
    'on',
    'after',
    'Trainer',
    'BaseComponent',
    'AttrsComponent',
    'StatefulAttrsComponent',
    'ComponentCheckError',
]



# -*- coding: utf-8 -*-
from typing import Any, Callable, Optional, Union

from simpletrainer import BaseComponent, DefaultSettings, Trainer, after, on
from simpletrainer.common.types import Batch, BatchOutput, MetricDict
from simpletrainer.integrations import is_torchmetrics_available
from simpletrainer.utils.common import pretty_repr, smartgetter

if is_torchmetrics_available():
    import torchmetrics as tm

    TorchMetric = Union[tm.Metric, tm.MetricCollection]


TargetGetter = Callable[[Batch], Any]
PredGetter = Callable[[BatchOutput], Any]


class TorchMetricsWrapper(BaseComponent):
    only_main_process = False

    def __init__(
        self,
        metric: 'TorchMetric',
        on_train: bool = True,
        on_valid: bool = True,
        pred_getter: Optional[PredGetter] = None,
        target_getter: Optional[TargetGetter] = None,
        compute_intervel: Optional[int] = None,
    ):
        if not is_torchmetrics_available():
            raise ImportError('torchmetrics is not available')

        if not on_train and not on_valid:
            raise ValueError('on_train and on_valid must be True at least one')

        if on_train:
            self.train_metric = metric.clone()
        if on_valid:
            self.valid_metric = metric.clone()

        self.on_train = on_train
        self.on_valid = on_valid

        self.pred_getter = pred_getter or smartgetter(DefaultSettings.pred_key)
        self.target_getter = target_getter or smartgetter(DefaultSettings.target_key)
        self.compute_intervel = compute_intervel

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        if self.compute_intervel is None:
            self.compute_intervel = self.suggest_intervel(trainer.num_steps_per_epoch)

    @on(Trainer.EVENT.START)
    def set_device(self, trainer: Trainer):
        if self.on_train:
            self.train_metric.to(trainer.device)
        if self.on_valid:
            self.valid_metric.to(trainer.device)

    @after(Trainer.run_train_batch)
    def update_train_metrics(self, trainer: Trainer):
        if self.on_train:
            self._update(trainer, self.train_metric)

    @after(Trainer.run_valid_batch)
    def update_valid_metrics(self, trainer: Trainer):
        if self.on_valid:
            self._update(trainer, self.train_metric)

    def _update(self, trainer: Trainer, metric: 'TorchMetric'):
        batch = trainer.state.batch
        batch_output = trainer.state.batch_output
        pred = self.pred_getter(batch_output)
        target = self.target_getter(batch)
        metric.update(pred, target)
        if (trainer.state.num_steps % self.compute_intervel) == 0:   # type: ignore
            trainer.update_stage_metrics(**self.compute(metric))

    @after(Trainer.run_stage)
    def reset_metrics(self, trainer: Trainer):
        if self.on_train:
            self.train_metric.reset()
        if self.on_valid:
            self.valid_metric.reset()

    @staticmethod
    def compute(metric: 'TorchMetric') -> MetricDict:
        score_dict = metric.compute()
        if not isinstance(score_dict, dict):
            score_dict = {metric.__class__.__name__: score_dict}
        score_dict = {k: v.item() for k, v in score_dict.items()}
        return score_dict

    def suggest_intervel(self, steps_per_epoch: Optional[int]) -> int:
        if steps_per_epoch is None:
            return 100
        else:
            return min((steps_per_epoch // 100) + 1, 100)

    def __repr__(self) -> str:
        fields = {}
        if self.on_train:
            fields['metric'] = repr(self.train_metric)
        else:
            fields['metric'] = repr(self.valid_metric)
        fields.update(
            on_train=self.on_train,
            on_valid=self.on_valid,
            compute_intervel=self.compute_intervel,
        )
        return pretty_repr(fields, self.__class__.__name__)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False



import logging
from typing import Literal, Optional

from torch.optim.lr_scheduler import _LRScheduler  # type: ignore

from simpletrainer import BaseComponent, Trainer, after
from simpletrainer.integrations import is_transformers_available

logger = logging.getLogger(__name__)
SchedulerTypeStr = Literal[
    'linear',
    'cosine',
    'cosine_with_restarts',
    'polynomial',
    'constant',
    'constant_with_warmup',
]


class TransformersLRScheduler(BaseComponent):
    only_main_process = False
    hyper_param_names = ['lr_scheduler_type', 'warmup_steps', 'max_warmup_steps']
    lr_scheduler: '_LRScheduler'

    def __init__(
        self,
        lr_scheduler_type: SchedulerTypeStr,
        warmup_steps: Optional[float] = None,
        total_steps: Optional[int] = None,
        max_warmup_steps: Optional[int] = None,
    ) -> None:
        super().__init__()
        if not is_transformers_available():
            raise ValueError('transformers is not installed')
        self.lr_scheduler_type = lr_scheduler_type
        self.warmup_steps = warmup_steps
        self.total_steps = total_steps
        self.max_warmup_steps = max_warmup_steps

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        total_steps_from_trainer = trainer.total_steps
        if self.total_steps is None:
            if total_steps_from_trainer is None:
                raise ValueError('Can not infer total steps from lr scheduler')
            else:
                logger.info(
                    'use trainer.total_steps (infer from dataloader), total_steps: %s',
                    total_steps_from_trainer,
                )
                total_steps = total_steps_from_trainer
        else:
            if self.total_steps != total_steps_from_trainer:
                logger.warning(
                    f'self.total_steps ({self.total_steps}) is not equal to \
                    trainer.total_steps ({total_steps_from_trainer})'
                )
            total_steps = int(self.total_steps)

        if self.warmup_steps is not None:
            if self.warmup_steps < 1:
                num_warmup_steps = int(self.warmup_steps * total_steps)
                if self.max_warmup_steps:
                    num_warmup_steps = min(num_warmup_steps, self.max_warmup_steps)
            else:
                num_warmup_steps = int(self.warmup_steps)
        else:
            num_warmup_steps = None
        self.warmup_steps = num_warmup_steps

        from transformers import get_scheduler

        self.lr_scheduler = get_scheduler(  # type: ignore
            name=self.lr_scheduler_type,
            optimizer=trainer.optimizer,
            num_warmup_steps=num_warmup_steps,
            num_training_steps=total_steps,
        )

    @after(Trainer.step, try_first=True)
    def step(self, trainer: Trainer) -> None:
        self.lr_scheduler.step()



# -*- coding: utf-8 -*-
import os
import warnings
from typing import Optional

import requests

from simpletrainer import define
from simpletrainer.components.notification.base import BaseNotification
from simpletrainer.components.notification.common import (
    NotificationInfo,
    NotificationType,
)

_notification_type_title_map = {
    NotificationType.START: ' 🎬 Train Started ',
    NotificationType.FINISH: '🎉 Train Completed',
    NotificationType.CRASH: '💥 Train Crashed',
}


@define(only_main_process=True, tags=('notification',))
class BarkNotification(BaseNotification):
    token: Optional[str] = None

    def __attrs_post_init__(self) -> None:
        self.token = self.token or os.getenv('BARK_TOKEN')
        if self.token is None:
            warnings.warn('Can not get bark token, BarkNotificationCallback will disabled!')
            self.bark_api = None
        else:
            self.bark_api = f'https://api.day.app/{self.token}/'

    def send(self, info: NotificationInfo) -> None:
        if self.bark_api is not None:
            requests.post(
                self.bark_api,
                json={
                    'title': _notification_type_title_map[info.type_],
                    'body': info.content,
                },
            )



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Optional, cast

import torch

from simpletrainer import AttrsComponent, Trainer, after, before, define, field, on
from simpletrainer.components.basic.metric_tracker import MetricTracker
from simpletrainer.utils.torch import (
    get_module_gradient_summary,
    get_module_learning_rate_summary,
    get_module_parameter_summary,
)


@define(only_main_process=True)
class ModelWatcher(AttrsComponent):
    interval: int = field(default=0, save=True)
    watch_metrics: bool = True
    watch_parameter: bool = False
    watch_gradient: bool = False
    watch_parameter_histogram: bool = False
    watch_learning_rate: bool = False
    watch_output_dir: bool = False
    metric_tracker: MetricTracker = field(init=False)

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        self.metric_tracker = trainer.hook_engine.setdefault(MetricTracker(trainer.core_metric))
        self.interval = self.interval or self._suggest_intervel(trainer.num_steps_per_epoch)

    def _suggest_intervel(self, steps_per_epoch: Optional[int]) -> int:
        if steps_per_epoch is None:
            return 100
        else:
            if self.watch_parameter or self.watch_gradient or self.watch_parameter_histogram:
                return min((steps_per_epoch // 20) + 1, 100)
            else:
                return min((steps_per_epoch // 100) + 1, 100)

    @before(Trainer.step)
    def _log_step(self, trainer: Trainer):
        self.interval = cast(int, self.interval)

        if trainer.cum_steps % self.interval != 0:
            return

        if self.watch_metrics:
            trainer.log(trainer.state.stage_metrics)

        if self.watch_learning_rate:
            lr_summary = get_module_learning_rate_summary(trainer.model, trainer.optimizer)
            trainer.log(lr_summary, prefix='learning_rate')

        if self.watch_parameter:
            param_mean, param_std = get_module_parameter_summary(trainer.model)
            trainer.log(param_mean, prefix='parameter_mean')
            trainer.log(param_std, prefix='parameter_std')

        if self.watch_gradient:
            grad_mean, grad_std = get_module_gradient_summary(trainer.model)
            trainer.log(grad_mean, prefix='gradient_mean')
            trainer.log(grad_std, prefix='gradient_std')

        if self.watch_parameter_histogram:
            parameters_to_log: dict[str, torch.Tensor] = {name: param for name, param in trainer.model.named_parameters()}
            trainer.log(parameters_to_log, prefix='parameter_histogram')

    @after(Trainer.run_valid_stage)
    def watch_valid_metrics(self, trainer: Trainer):
        validation_metrics = trainer.valid_metrics_history[-1]
        trainer.log(validation_metrics, prefix='validation', step=trainer.current_epoch)

    @on(Trainer.EVENT.FINISH, try_last=True)
    def watch_output(self, trainer) -> None:
        if self.watch_output_dir:
            trainer.logger.log_artifact(str(trainer.output_dir))

    @on(Trainer.EVENT.FINISH)
    def watch_best_metrics(self, trainer: Trainer):
        if trainer.do_eval and self.metric_tracker.best_score is not None:
            best_metrics = trainer.valid_metrics_history[self.metric_tracker.best_epoch - 1]
        else:
            best_metrics = trainer.train_metrics_history[-1]
        if self.watch_metrics:
            trainer.logger.log_summary(trainer.hyper_params, best_metrics)



from pathlib import Path
from typing import Any, Callable

import torch

from simpletrainer import AttrsComponent, DefaultSettings, Trainer, define, field, on
from simpletrainer.common.sign_metric import SignMetric
from simpletrainer.components.basic.best_model_state_tracker import (
    BestModelStateTracker,
)

SaveFunction = Callable[[Any, Path], None]


def torch_save(obj: Any, path: Path) -> None:
    torch.save(obj, path)


@define(only_main_process=True)
class SaveBestModel(AttrsComponent):
    sign_metric: SignMetric = field(default=None, converter=SignMetric.from_str)
    save_state_dict: bool = True
    save_model: bool = False
    save_function: SaveFunction = field(default=torch_save)

    def __attrs_post_init__(self) -> None:
        if (not self.save_state_dict) and (not self.save_model):
            raise ValueError('Either save_state_dict or save_model should be True')

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        self.sign_metric = self.sign_metric or trainer.core_metric
        self.best_model_tracker = trainer.hook_engine.setdefault(BestModelStateTracker(sign_metric=self.sign_metric))

    @on(Trainer.EVENT.FINISH)
    def track_best_model(self, trainer: Trainer) -> None:
        if self.save_state_dict:
            self.save_function(
                self.best_model_tracker.best_model_state, trainer.output_dir / DefaultSettings.best_model_state_file_name
            )

        if self.save_model:
            trainer.raw_model.load_state_dict(self.best_model_tracker.best_model_state)
            self.save_function(trainer.raw_model, trainer.output_dir / DefaultSettings.best_model_file_name)



# -*- coding: utf-8 -*-
from simpletrainer import BaseComponent, Trainer, on
from simpletrainer.components.notification.common import (
    NotificationInfo,
    crash_info,
    finish_info,
    start_info,
)


class BaseNotification(BaseComponent):
    @on(Trainer.EVENT.START)
    def send_start_info(self, trainer: Trainer) -> None:
        self.send(start_info(trainer))

    @on(Trainer.EVENT.FINISH)
    def send_finish_info(self, trainer: Trainer) -> None:
        self.send(finish_info(trainer))

    @on(Trainer.EVENT.CRASH)
    def send_crash_info(self, trainer: Trainer) -> None:
        self.send(crash_info(trainer))

    def send(self, info: NotificationInfo) -> None:
        raise NotImplementedError



from typing import Any, Callable, TypeVar

T = TypeVar('T')
Oprator = Callable[[T, Any], T]


def replace(value: T, other_value: T) -> T:
    return other_value


def dict_update(value: dict, other_value: dict) -> dict:
    value.update(other_value)
    return value


def or_(value: bool, other_value: bool) -> bool:
    return value or other_value


def and_(value: bool, other_value: bool) -> bool:
    return value and other_value


class Opration:
    def __init__(self, oprator: Oprator, value: Any):
        self.oprator = oprator
        self.value = value

    def __call__(self, value: T) -> T:
        return self.oprator(value, self.value)



from simpletrainer import BaseComponent, Trainer, after


class SaveCheckpoint(BaseComponent):
    def __init__(self, save_every_steps: int = 100) -> None:
        super().__init__()
        self.save_every_steps = save_every_steps

    @after(Trainer.run_epoch)
    def save_checkpoint(self, trainer: Trainer) -> None:
        trainer.save(f'epoch-{trainer.current_epoch}')

    def __repr__(self) -> str:
        return 'SaveCheckpoint()'



# -*- coding: utf-8 -*-
import simpletrainer.common.oprator as op
from simpletrainer import BaseComponent, Trainer, after


class TryOverfit(BaseComponent):
    only_main_process = False

    def post_init_with_trainer(self, trainer: Trainer):
        self.batch = next(iter(trainer.train_dataloader))

    @after(Trainer.generate_batch, try_first=True)
    def replace_batch(self, trainer: Trainer):
        return op.Opration(op.replace, self.batch)

    def __repr__(self):
        return 'TryOverfit()'



# -*- coding: utf-8 -*-
import logging
from typing import Optional, Union

from simpletrainer import AttrsComponent, Trainer, after, define, field, op
from simpletrainer.common.sign_metric import SignMetric
from simpletrainer.components.basic import MetricTracker

logger = logging.getLogger(__name__)


@define
class EarlyStopping(AttrsComponent):
    patience: int = field(default=1, hyper_param=True)
    sign_metric: SignMetric = field(default=None, converter=SignMetric.from_str)
    metric_tracker: MetricTracker = field(init=False)

    def __init__(
        self,
        patience: int = 1,
        sign_metric: Optional[Union[str, SignMetric]] = None,
    ) -> None:
        if patience <= 0:
            raise ValueError('patience must be greater than 0')

        self.patience = patience
        self._init_sign_metric = sign_metric

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        sign_metric = self._init_sign_metric or trainer.core_metric
        if isinstance(sign_metric, str):
            sign_metric = SignMetric.from_str(sign_metric)
        self.sign_metric = sign_metric
        logger.info(f'component EarlyStopping track {self._init_sign_metric} with patience {self.patience}')
        self.metric_tracker = trainer.hook_engine.setdefault(MetricTracker(self.sign_metric))

    @after(Trainer.should_stop_loop)
    def early_stop(self, trainer: Trainer):
        no_patience = (trainer.current_epoch - self.metric_tracker.best_epoch) > self.patience
        if no_patience:
            logger.info(
                f'early stop because no patience, best epoch {self.metric_tracker.best_epoch}, current epoch {trainer.current_epoch}'
            )
        return op.Opration(op.or_, no_patience)



from typing import Optional

from attrs.converters import optional

from simpletrainer import StatefulAttrsComponent, Trainer, after, define, field
from simpletrainer.common.sign_metric import SignMetric


@define
class MetricTracker(StatefulAttrsComponent):
    sign_metric: SignMetric = field(default=None, converter=optional(SignMetric.from_str))
    train: bool = False
    best_epoch: int = field(default=0, save=True)
    best_score: Optional[float] = field(default=None, save=True)

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        self.sign_metric = self.sign_metric or trainer.core_metric

    @after(Trainer.run_stage)
    def track_metric(self, trainer: Trainer):
        if self.train != trainer.in_train_stage:
            return

        if self.train:
            metrics = trainer.latest_train_metrics
        else:
            metrics = trainer.latest_valid_metrics

        if metrics is None:
            raise ValueError(f'Metrics are not available')

        score = metrics.get(self.sign_metric.name)
        if score is None:
            raise ValueError(f'Metric {self.sign_metric.name} not found in {metrics}')

        if self.best_score is None:
            self.best_score = score
            self.best_epoch = trainer.current_epoch
        elif self.sign_metric.is_better(score, self.best_score):
            self.best_score = score
            self.best_epoch = trainer.current_epoch



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Mapping, Union

import torch
import torch.utils.data
from typing_extensions import TypeAlias

Batch: TypeAlias = Mapping[str, Any]
BatchOutput: TypeAlias = Mapping[str, Any]
MetricDict: TypeAlias = Dict[str, float]
Dataset: TypeAlias = torch.utils.data.Dataset[Batch]
LRScheduler: TypeAlias = torch.optim.lr_scheduler._LRScheduler
Prime: TypeAlias = Union[str, float, int, bool, None]
JSON: TypeAlias = Union[Prime, List['JSON'], Dict[str, 'JSON']]
HyperParams: TypeAlias = Dict[str, Prime]
Device: TypeAlias = Union[str, torch.device]
PathOrStr: TypeAlias = Union[str, Path]



from simpletrainer.loggers.base import DeepLearningLogger
from simpletrainer.loggers.mlflow import MlflowLogger
from simpletrainer.loggers.registry import DeepLearningLoggerRegistry
from simpletrainer.loggers.tensorboard import TensorboardLogger
from simpletrainer.loggers.wandb import WandbLogger

__all__ = [
    'DeepLearningLogger',
    'MlflowLogger',
    'TensorboardLogger',
    'WandbLogger',
    'DeepLearningLoggerRegistry',
]



from simpletrainer.components.basic.best_model_state_tracker import (
    BestModelStateTracker,
)
from simpletrainer.components.basic.file_handler import FileHandler
from simpletrainer.components.basic.metric_tracker import MetricTracker
from simpletrainer.components.basic.model_watcher import ModelWatcher
from simpletrainer.components.basic.save_best_model import SaveBestModel
from simpletrainer.components.basic.save_checkpoint import SaveCheckpoint
from simpletrainer.components.basic.scaler_monitor import ScalerMonitor
from simpletrainer.components.basic.timer import Timer

__all__ = [
    'BestModelStateTracker',
    'FileHandler',
    'MetricTracker',
    'SaveBestModel',
    'ScalerMonitor',
    'Timer',
    'ModelWatcher',
    'SaveCheckpoint',
]



from __future__ import annotations

import inspect
import logging
import random
import shutil
import time
from contextlib import contextmanager
from pathlib import Path
from types import MethodType
from typing import Any, Callable, Mapping, Optional, Union

from coolname import generate_slug


@contextmanager
def temp_random_seed(seed: int):
    old_state = random.getstate()
    random.seed(seed)
    yield
    random.setstate(old_state)


def random_experiment_name() -> str:  # type: ignore
    seed = int(time.time() * 1000)
    with temp_random_seed(seed):
        return generate_slug(2)


def remove_dir(directory: Union[Path, str]) -> None:
    directory = Path(directory)

    if not directory.exists():
        return

    shutil.rmtree(directory)


def get_init_params(obj: Any) -> dict[str, Any]:
    if obj.__class__.__init__ == object.__init__:
        return {}

    init_signature = inspect.signature(obj.__class__.__init__)
    param_names = [
        p for p in init_signature.parameters.values() if p.name != 'self' and p.kind not in (p.VAR_KEYWORD, p.VAR_POSITIONAL)
    ]
    param_names = sorted([p.name for p in param_names])
    params = {}
    for name in param_names:
        try:
            value = getattr(obj, name)
        except AttributeError:
            raise AttributeError(
                f'{obj.__class__.__name__} has no attribute {name},maybe attribute name is different from the __init__ signature'
            )
        params[name] = value
    return params


def pretty_repr(fields: Mapping[str, Any], name: Optional[str] = None) -> str:
    name = name or ''
    format_string = name + '('
    if not fields:
        format_string += ')'
    else:
        for k, v in fields.items():
            format_string += '\n'
            format_string += f'    {k}: {v}'
        format_string += '\n)'
    return format_string


def pretty_obj_str(obj: Any, init: bool = False) -> str:
    fields = get_init_params(obj) if init else obj.__dict__
    name = obj.__class__.__name__
    return pretty_repr(fields, name)


def monkey_patch_method(obj, method: Callable) -> None:
    setattr(obj, method.__name__, MethodType(method, obj))


def add_file_handler(log_file: Union[str, Path], level: int = logging.INFO):
    file_handler = logging.FileHandler(str(log_file))
    formatter = logging.Formatter('%(asctime)s - %(message)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)
    file_handler.setLevel(level)
    logging.root.addHandler(file_handler)


def smartget(obj: Any, name: str, default: Any = None) -> Any:
    if hasattr(obj, name):
        return getattr(obj, name)
    if hasattr(obj, '__getitem__'):
        return obj[name]
    return default


class smartgetter:
    def __init__(self, name: str, default: Any = None) -> None:
        self.name = name
        self.default = default

    def __call__(self, obj: Any) -> Any:
        return smartget(obj, self.name, self.default)



from typing import Callable, Union, cast

from simpletrainer import AttrsComponent, Trainer, after, define

ScalerGetter = Callable[[Trainer], Union[int, float]]


@define
class ScalerMonitor(AttrsComponent):
    name: str
    getter: ScalerGetter
    intervel: int

    @after(Trainer.run_batch)
    def collect(self, trainer: Trainer):
        self.intervel = cast(int, self.intervel)

        if trainer.state.num_batches % self.intervel == 0:
            trainer.update_stage_metrics(**{self.name: self.getter(trainer)})



# -*- coding: utf-8 -*-
from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any, ClassVar

import torch

from simpletrainer.common.types import HyperParams, MetricDict

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer


class DeepLearningLogger(ABC):
    name: ClassVar[str]

    def post_init(self, trainer: 'Trainer') -> None:
        pass

    def log_metrics(self, metrics: dict[str, float], step: int | None = None) -> None:
        raise NotImplementedError

    def log_tensors(self, tensors: dict[str, torch.Tensor], step: int | None = None) -> None:
        raise NotImplementedError

    def log_summary(self, hyper_params: HyperParams, metrics: MetricDict) -> None:
        raise NotImplementedError

    def log_artifact(self, artifact: str) -> None:
        raise NotImplementedError

    @staticmethod
    def add_prefix(prefix: str, data: dict[str, Any]) -> dict[str, Any]:
        return {f'{prefix}/{key}': value for key, value in data.items()}

    def teardown(self) -> None:
        pass



# -*- coding: utf-8 -*-
import inspect
import logging
from typing import Callable

import torch

from simpletrainer import AttrsComponent, Trainer, define, field, on

logger = logging.getLogger(__name__)
CrossEntropyGetter = Callable[[Trainer], torch.nn.CrossEntropyLoss]


def auto_find_cross_entropy(trainer: Trainer):
    for _, v in inspect.getmembers(trainer.model):
        if isinstance(v, torch.nn.CrossEntropyLoss):
            return v
    raise ValueError('No CrossEntropyLoss found in model')


@define
class LabelSmoothing(AttrsComponent):
    rate: float = field(default=0.1, hyper_param=True)
    cross_entropy_getter: CrossEntropyGetter = auto_find_cross_entropy

    @on(Trainer.EVENT.START)
    def set_label_smoothing(self, trainer: Trainer):
        try:
            loss_instance = self.cross_entropy_getter(trainer)
            loss_instance.label_smoothing = self.rate
            logging.info(
                'get cross entropy loss success, set label smoothing {}',
                self.rate,
            )
        except Exception:
            logging.warning('get cross entropy loss failed, LabelSmoothing disabled')



# -*- coding: utf-8 -*-
from typing import Mapping

import torch

from simpletrainer import AttrsComponent, Trainer, after, define, field
from simpletrainer.common.sign_metric import SignMetric
from simpletrainer.components.basic.metric_tracker import MetricTracker


@define(only_main_process=True)
class BestModelStateTracker(AttrsComponent):
    sign_metric: SignMetric = field(default=None, converter=SignMetric.from_str)
    best_model_state: Mapping[str, torch.Tensor] = field(init=False, export=True)
    metric_tracker: MetricTracker = field(init=False)

    def post_init_with_trainer(self, trainer: Trainer) -> None:
        self.sign_metric = self.sign_metric or trainer.core_metric
        self.metric_tracker = trainer.hook_engine.setdefault(MetricTracker(sign_metric=self.sign_metric))

    @after(Trainer.run_epoch)
    def track_best_model(self, trainer: Trainer) -> None:
        if self.metric_tracker.best_epoch == trainer.current_epoch:
            self.best_model_state = trainer.model.state_dict()



# -*- coding: utf-8 -*-
import re

import torch

from simpletrainer import AttrsComponent, Trainer, after, define, field
from simpletrainer.utils.torch import get_params_with_pattern


@define(tags=('adversarial',))
class FGM(AttrsComponent):
    pattern: re.Pattern = field(converter=re.compile, hyper_param=True)
    adv_weight: float = field(default=1.0, hyper_param=True)
    eps: float = field(default=1e-8, hyper_param=True)

    @after(Trainer.set_loss, try_first=True)
    def add_adv(self, trainer: Trainer):
        if trainer.in_train_stage:
            return

        loss = trainer.state.loss
        params = get_params_with_pattern(trainer.model, self.pattern)

        grad_params = torch.autograd.grad(outputs=loss, inputs=params, create_graph=True)
        params_noise = [self.calculate_noise(grad) for grad in grad_params]

        for param, noise in zip(params, params_noise):
            param.data.add_(noise)

        fgm_loss = self.adv_weight * trainer.model(**trainer.state.batch)['loss']
        fgm_loss.backward()
        trainer.update_stage_metrics(fgm_loss=fgm_loss.item())

        for param, noise in zip(params, params_noise):
            param.data.sub_(noise)

    def calculate_noise(self, grad: torch.Tensor) -> torch.Tensor:
        return self.adv_weight * (grad / (grad.pow(2).sum() + self.eps))



import tqdm

from simpletrainer import BaseComponent, Trainer, after, before
from simpletrainer.common.types import MetricDict


class TqdmProgressBar(BaseComponent):
    only_main_process = True
    tags = ['progress_bar']

    def __init__(self, update_intervel: int = 10) -> None:
        self.update_intervel = update_intervel

    @before(Trainer.run_stage)
    def start_train(self, trainer: Trainer):
        if trainer.in_train_stage:
            total = trainer.train_data_info.num_batches_per_epoch
        else:
            total = trainer.valid_data_info.num_batches_per_epoch  # type: ignore

        self.progress_bar = tqdm.tqdm(
            total=total,
            desc=self.get_trainer_description(trainer),
            unit='bat',
            initial=trainer.state.num_batches,
        )

    @after(Trainer.run_stage, try_first=True)
    def close(self, trainer: Trainer):
        self.progress_bar.close()

    @after(Trainer.run_batch)
    def advance_batch(self, trainer: Trainer):
        self.progress_bar.update(1)
        if trainer.state.num_batches % self.update_intervel == 1:
            description = (
                self.get_trainer_description(trainer) + ' | ' + self.get_metrics_description(trainer.state.stage_metrics)
            )
            self.progress_bar.set_description(description)

    @staticmethod
    def get_trainer_description(trainer: Trainer) -> str:
        stage_name = 'Trian' if trainer.in_train_stage else 'Valid'
        description = f'Epoch {trainer.current_epoch}/{trainer.epochs} | {stage_name}'
        return description

    @staticmethod
    def get_metrics_description(metrics: MetricDict) -> str:
        processed_metrics = {}
        for k, v in metrics.items():
            if isinstance(v, float):
                processed_metrics[k] = f'{v:.3f}'
            else:
                processed_metrics[k] = v
        description = ' | '.join(f'{k}: {v}' for k, v in processed_metrics.items())
        return description

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
from datetime import datetime

from simpletrainer import AttrsComponent, DefaultSettings, Trainer, define, field, on

logger = logging.getLogger(__name__)


@define(only_main_process=True)
class Timer(AttrsComponent):
    start_time: datetime = field(init=False, export=True)
    end_time: datetime = field(init=False, export=True)

    @on(Trainer.EVENT.START, try_first=True)
    def start(self, trainer: Trainer):
        self.start_time = datetime.now()

    @on(Trainer.EVENT.FINISH, try_first=True)
    def stop(self, trainer: Trainer):
        self.end_time = datetime.now()
        logger.info(f'🎉  {trainer.experiment_name}/{trainer.run_name} Finish, Elapsed Time: {self.elapsed}')

    @property
    def start_string(self) -> str:
        return self.start_time.strftime(DefaultSettings.date_format)

    @property
    def end_string(self) -> str:
        return self.end_time.strftime(DefaultSettings.date_format)

    @property
    def elapsed(self) -> str:
        time_delta = self.end_time - self.start_time
        mins = time_delta.total_seconds() // 60
        elapsed = f'{int(mins // 60)}h,{int(mins % 60)}m'
        return elapsed



from __future__ import annotations

import inspect
import logging
from collections import defaultdict
from enum import Enum
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)
from warnings import warn

from simpletrainer.common.oprator import Opration
from simpletrainer.core.component import BaseComponent

if TYPE_CHECKING:
    from simpletrainer.core.trainer import Trainer

logger = logging.getLogger(__name__)

T = TypeVar('T')
CT = TypeVar('CT', bound=BaseComponent)
OnOrBeforeMethod = Callable[[Any, 'Trainer'], None]
AfterMethod = Callable[[Any, 'Trainer'], Optional[Opration]]
EntryPoint = Union[Callable, 'TrainerEvent']


class Preposition(str, Enum):
    before = 'before'
    after = 'after'
    on = 'on'


class TrainerEvent(str, Enum):
    START = 'START'
    CRASH = 'CRASH'
    FINISH = 'FINISH'
    TEARDOWN = 'TEARDOWN'


def entrypoint(method):
    @wraps(method)
    def wrapped_method(self: 'Trainer', *args, **kwargs):
        self.state.entrypoint_stack.append(method.__name__)
        self.hook_engine.before(method)
        raw_return = method(self, *args, **kwargs)
        final_return = self.hook_engine.after(method, raw_return)
        self.state.entrypoint_stack.pop(-1)
        return final_return

    return wrapped_method


def get_entrypoint_name(entrypoint: EntryPoint) -> str:
    if isinstance(EntryPoint, TrainerEvent):
        return entrypoint.value
    elif isinstance(entrypoint, str):
        return entrypoint
    else:
        return entrypoint.__name__


@runtime_checkable
class TrainerHook(Protocol):
    __qualname__: str
    _preposition: Preposition
    _entrypoint_name: str
    _priority: int

    def __call__(self, trainer: Trainer) -> Optional[Opration]:
        ...


class HookCollection:
    def __init__(self) -> None:
        self.data: dict[Preposition, list[TrainerHook]] = {
            Preposition.before: [],
            Preposition.after: [],
            Preposition.on: [],
        }

    def add(self, method: TrainerHook) -> None:
        self.data[method._preposition].append(method)
        self.data[method._preposition].sort(key=lambda method: method._priority)


class TrainerHookEngine:
    def __init__(self, trainer: 'Trainer') -> None:
        # Map: EntryPoint -> HookCollection
        self.entrypoint_hooks: dict[str, HookCollection] = defaultdict(lambda: HookCollection())
        self.trainer = trainer

    def _register(self, component: BaseComponent) -> None:
        hooks = self.inspect_trainer_hook(component)
        if not hooks:
            logger.warning(f'{component} do not have any TrainerHook')

        for hook in hooks:
            self.entrypoint_hooks[hook._entrypoint_name].add(hook)

    def _add(self, component: BaseComponent) -> None:
        logger.debug(f'Add component {component.__class__.__name__}')
        self.trainer._components.append(component)
        self.refresh()

    @staticmethod
    def inspect_trainer_hook(obj: Any) -> list[TrainerHook]:
        trainer_hooks: list[TrainerHook] = []
        for _, member in inspect.getmembers(obj):
            is_method_or_func = inspect.ismethod(member) or inspect.isfunction(member)
            if is_method_or_func and isinstance(member, TrainerHook):
                trainer_hooks.append(member)
        return trainer_hooks

    def register(self, component: BaseComponent) -> None:
        for other_component in self.trainer._components:
            if component == other_component:
                warn(f'{component} already exists')
                return

        for other_component in self.trainer._components:
            component.check_compatibility(other_component)

        if self.trainer.is_main_process or (not component.only_main_process):
            self._add(component)

    def setdefault(self, component: CT) -> CT:
        component.post_init_with_trainer(self.trainer)

        for other_component in self.trainer._components:
            if component == other_component:
                return other_component  # type: ignore

        for other_component in self.trainer._components:
            component.check_compatibility(other_component)

        if self.trainer.is_main_process or (not component.only_main_process):
            self._add(component)

        return component

    def pop(self, component: CT) -> CT:
        output_component = self.trainer._components.pop(self.trainer._components.index(component))
        self.refresh()
        return output_component  # type: ignore

    def find(self, component_class: Type[CT], **attrs) -> Optional[CT]:
        for component in self.trainer._components:
            if not isinstance(component, component_class):
                continue

            if all(getattr(component, k) == v for k, v in attrs.items()):
                return component

    def refresh(self) -> None:
        self.entrypoint_hooks = defaultdict(lambda: HookCollection())
        for component in self.trainer._components:
            self._register(component)

    def after(self, entrypoint: EntryPoint, raw_return: T) -> T:
        entrypoint_name = get_entrypoint_name(entrypoint)
        if entrypoint_name not in self.entrypoint_hooks:
            return raw_return

        for hook in self.entrypoint_hooks[entrypoint_name].data[Preposition.after]:
            try:
                opration = hook(self.trainer)
                if opration is not None:
                    new_return = opration(raw_return)
                    assert type(new_return) == type(raw_return)
                    raw_return = new_return
            except Exception as e:
                e.args = (f'{hook.__qualname__}', *e.args)
                raise
        return raw_return

    def before(self, entrypoint: EntryPoint) -> None:
        entrypoint_name = get_entrypoint_name(entrypoint)
        if entrypoint_name not in self.entrypoint_hooks:
            return

        for hook in self.entrypoint_hooks[entrypoint_name].data[Preposition.before]:
            try:
                hook(self.trainer)
            except Exception as e:
                e.args = (f'{hook.__qualname__}', *e.args)
                raise

    def on(self, entrypoint: EntryPoint) -> None:
        entrypoint_name = get_entrypoint_name(entrypoint)
        if entrypoint_name not in self.entrypoint_hooks:
            return

        for hook in self.entrypoint_hooks[entrypoint_name].data[Preposition.on]:
            try:
                hook(self.trainer)
            except Exception as e:
                e.args = (f'{hook.__qualname__}', *e.args)
                raise


def _get_priority(try_first: bool = False, try_last: bool = False) -> int:
    if try_first and try_last:
        raise ValueError("Can't try first and last at the same time")
    priority = 50
    if try_first:
        priority = 0
    if try_last:
        priority = 100
    return priority


def after(
    entrypoint: EntryPoint,
    try_first: bool = False,
    try_last: bool = False,
    priority: Optional[int] = None,
) -> Callable[[AfterMethod], TrainerHook]:
    entrypoint_name = get_entrypoint_name(entrypoint)

    if priority is None:
        priority = _get_priority(try_first, try_last)
    elif priority < 0 or priority > 100:
        raise ValueError('Priority must be between 0 and 100')

    def bounded_target(target: AfterMethod) -> TrainerHook:
        target._entrypoint_name = entrypoint_name
        target._preposition = Preposition.after
        target._priority = priority
        return target  # type: ignore

    return bounded_target


def before(
    entrypoint: EntryPoint,
    try_first: bool = False,
    try_last: bool = False,
    priority: Optional[int] = None,
) -> Callable[[OnOrBeforeMethod], TrainerHook]:
    entrypoint_name = get_entrypoint_name(entrypoint)

    if priority is None:
        priority = _get_priority(try_first, try_last)
    elif priority < 0 or priority > 100:
        raise ValueError('Priority must be between 0 and 100')

    def bounded_target(target: OnOrBeforeMethod) -> TrainerHook:
        target._entrypoint_name = entrypoint_name
        target._preposition = Preposition.before
        target._priority = priority
        return target  # type: ignore

    return bounded_target


def on(
    entrypoint: TrainerEvent,
    try_first: bool = False,
    try_last: bool = False,
    priority: Optional[int] = None,
) -> Callable[[OnOrBeforeMethod], TrainerHook]:
    if priority is None:
        priority = _get_priority(try_first, try_last)
    elif priority < 0 or priority > 100:
        raise ValueError('Priority must be between 0 and 100')

    def bounded_target(target: OnOrBeforeMethod) -> TrainerHook:
        target._entrypoint_name = entrypoint.value
        target._preposition = Preposition.on
        target._priority = priority
        return target  # type: ignore

    return bounded_target



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['simpletrainer',
 'simpletrainer.common',
 'simpletrainer.components',
 'simpletrainer.components.adversarial',
 'simpletrainer.components.basic',
 'simpletrainer.components.debug',
 'simpletrainer.components.grad',
 'simpletrainer.components.lr_scheduler',
 'simpletrainer.components.metric',
 'simpletrainer.components.notification',
 'simpletrainer.components.terminal',
 'simpletrainer.core',
 'simpletrainer.loggers',
 'simpletrainer.utils']

package_data = \
{'': ['*']}

install_requires = \
['accelerate>=0.12.0,<0.13.0',
 'coolname>=1.1.0,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.5.1,<13.0.0',
 'tensorboard>=2.10.0,<3.0.0',
 'torchinfo>=1.7.0,<2.0.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'simpletrainer',
    'version': '0.3.2',
    'description': 'Simple PyTorch Trainer',
    'long_description': '# SimpleTrainer\n\nSimple PyTorch Trianer\n',
    'author': 'wangyuxin',
    'author_email': 'wangyuxin@mokahr.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import os
import warnings

import requests

from simpletrainer import define, field
from simpletrainer.components.notification.base import BaseNotification
from simpletrainer.components.notification.common import (
    NotificationInfo,
    NotificationType,
)

_notification_type_title_map = {
    NotificationType.START: 'Your train has started 🎬',
    NotificationType.FINISH: 'Your train is complete 🎉',
    NotificationType.CRASH: 'Your train is crashed 💥',
}


@define(only_main_process=True, tags=('notification',))
class WechatNotification(BaseNotification):
    webhook_url: str = field(default=None)

    def __attrs_post_init__(self):
        self.webhook_url = self.webhook_url or os.getenv('WECHAT_WEBHOOK_URL', '')

        if not self.webhook_url:
            warnings.warn('Can not get notification webhook url, WechatNotificationCallback will disabled!')

    def send(self, info: NotificationInfo) -> None:
        title = _notification_type_title_map[info.type_]
        msg = {
            'msgtype': 'text',
            'text': {'content': ''},
        }
        msg['text']['content'] = title + '\n\n' + info.content
        if self.webhook_url:
            requests.post(self.webhook_url, json=msg)



from dataclasses import dataclass
from typing import Literal


@dataclass
class SignMetric:
    sign: Literal['+', '-']
    name: str

    @classmethod
    def from_str(cls, sign_metric_str: str) -> 'SignMetric':
        if isinstance(sign_metric_str, SignMetric):
            return sign_metric_str

        sign = sign_metric_str[0]
        name = sign_metric_str[1:]
        if sign != '+' and sign != '-':
            raise ValueError(f'Invalid sign: {sign_metric_str[0]}')
        if len(name.strip()) == 0:
            raise ValueError(f'Empty metric, {sign_metric_str}')
        return cls(sign, name)

    def is_better(self, score_a: float, score_b: float) -> bool:
        if self.sign == '+':
            return score_a > score_b
        else:
            return score_a < score_b

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, SignMetric):
            return self.name == other.name
        else:
            return False

    def __repr__(self):
        return f'{self.sign}{self.name}'

    def __str__(self) -> str:
        return self.__repr__()



from simpletrainer.components.adversarial import FGM
from simpletrainer.components.basic import (
    BestModelStateTracker,
    FileHandler,
    MetricTracker,
    ModelWatcher,
    SaveBestModel,
    SaveCheckpoint,
    ScalerMonitor,
    Timer,
)
from simpletrainer.components.debug import TryOverfit
from simpletrainer.components.early_stopping import EarlyStopping
from simpletrainer.components.grad import GradClip, GradPenalty
from simpletrainer.components.label_smoothing import LabelSmoothing
from simpletrainer.components.lr_scheduler import TransformersLRScheduler
from simpletrainer.components.metric import TorchMetricsWrapper
from simpletrainer.components.notification import BarkNotification, WechatNotification
from simpletrainer.components.terminal import (
    RichInspect,
    RichProgressBar,
    TqdmProgressBar,
)

__all__ = [
    'ModelWatcher',
    'MetricTracker',
    'BestModelStateTracker',
    'Timer',
    'SaveCheckpoint',
    'SaveBestModel',
    'FileHandler',
    'ScalerMonitor',
    'TqdmProgressBar',
    'RichInspect',
    'RichProgressBar',
    'GradClip',
    'GradPenalty',
    'EarlyStopping',
    'TransformersLRScheduler',
    'LabelSmoothing',
    'FGM',
    'TorchMetricsWrapper',
    'TryOverfit',
    'WechatNotification',
    'BarkNotification',
]



from simpletrainer.components.notification.bark import BarkNotification
from simpletrainer.components.notification.wechat import WechatNotification

__all__ = ['BarkNotification', 'WechatNotification']
